"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __rest = (this && this.__rest) || function (s, e) {
    var t = {};
    for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p) && e.indexOf(p) < 0)
        t[p] = s[p];
    if (s != null && typeof Object.getOwnPropertySymbols === "function")
        for (var i = 0, p = Object.getOwnPropertySymbols(s); i < p.length; i++) {
            if (e.indexOf(p[i]) < 0 && Object.prototype.propertyIsEnumerable.call(s, p[i]))
                t[p[i]] = s[p[i]];
        }
    return t;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.MethodHandler = void 0;
const lodash_1 = require("lodash");
const ajv_1 = __importDefault(require("ajv"));
const http_errors_1 = __importDefault(require("http-errors"));
const http_status_codes_1 = __importDefault(require("http-status-codes"));
const ajv_utils_1 = require("../ajv-utils");
const method_result_1 = require("../method-result");
const promise_utils_1 = require("../promise-utils");
const ajv = new ajv_1.default({ coerceTypes: true, useDefaults: true });
class MethodHandler {
    constructor(_method, _methodMetadata, _controller) {
        this._method = _method;
        this._methodMetadata = _methodMetadata;
        this._controller = _controller;
        // Pre-bind the handleRequest function so we can send it to express routes.
        this.handleRequest = this.handleRequest.bind(this);
        // Build the request validator.
        const requestSchema = lodash_1.get(_methodMetadata, ["request", "schema"]);
        this._requestValidator = requestSchema
            ? ajv.compile(requestSchema)
            : ajv_utils_1.NoOpAjvValidator;
        // Build the response validators.
        //  These are mapped based on status codes.
        this._responseValidators = lodash_1.mapValues(_methodMetadata.responses, (response) => {
            if (response.schema) {
                return ajv.compile(response.schema);
            }
            else {
                return ajv_utils_1.NoOpAjvValidator;
            }
        });
        // Build validators for query params.
        this._queryValidators = lodash_1.mapValues(_methodMetadata.queryParams, createQueryParamValidator);
        // Build validators for path params.
        this._pathValidators = lodash_1.mapValues(_methodMetadata.pathParams, createPathParamValidator);
    }
    /**
     * Handle an express request.
     *
     * This method is pre-bound.  Usage of `this` is safe when
     * this method is passed to the router.
     *
     * This method is async to assist with testing.  Its promise does
     * not need to be consumed, and errors will be passed to the
     * express `next` function.
     */
    handleRequest(req, res, next) {
        return __awaiter(this, void 0, void 0, function* () {
            // Catch any errors and pass them to express.
            yield this._executeRequest(req, res).catch(next);
        });
    }
    _executeRequest(req, res) {
        var _a;
        return __awaiter(this, void 0, void 0, function* () {
            // Ensure the request is valid.
            this._validateRequest(req);
            // Collect the arguments for the method based on method arg decorators.
            const args = this._collectMethodArgs(req, res);
            // Execute the method to handle the request.
            const methodResult = this._method.apply(this._controller, args);
            // The method may or may not have returned a promise.  Await it if so.
            let result = yield promise_utils_1.maybeAwaitPromise(methodResult);
            if (result === undefined) {
                // Throw an error to the user.  Express will return this into a 500.
                throw new Error("Controller methods must return a result.");
            }
            let statusCode = 200;
            let headers = {};
            let cookies = {};
            let body = result;
            let contentType = undefined;
            let sendRaw = false;
            if (result instanceof method_result_1.ResultBuilder) {
                if (result.handled) {
                    return;
                }
                statusCode = result.statusCode;
                headers = result.headers;
                cookies = result.cookies;
                contentType = result.contentType;
                sendRaw = result.raw;
                body = result.body;
            }
            const hasBody = result !== undefined;
            const headerContentType = (_a = headers["Content-Type"]) !== null && _a !== void 0 ? _a : headers["content-type"];
            if (hasBody) {
                // Ensure the response matches the documented response.
                this._validateResponse(statusCode, body);
                if (contentType === undefined) {
                    // Assume body is json by default.
                    // text/plain might make more sense for strings, but this preserves backwards compatibility.
                    contentType = headerContentType !== null && headerContentType !== void 0 ? headerContentType : "application/json";
                }
            }
            if (contentType != null && headerContentType == null) {
                headers["Content-Type"] = contentType;
            }
            // Set any headers that were requested.
            for (const key of Object.keys(headers)) {
                res.setHeader(key, headers[key]);
            }
            // Set requested cookies.
            for (const key of Object.keys(cookies)) {
                const _b = cookies[key], { value } = _b, cookieSettings = __rest(_b, ["value"]);
                res.cookie(key, value, cookieSettings);
            }
            // Set the status code.
            res.status(statusCode);
            if (hasBody) {
                if (sendRaw) {
                    res.send(body);
                }
                else if (contentType === "application/json" ||
                    typeof body === "object") {
                    res.json(body);
                }
                else {
                    res.send(body);
                }
            }
            else {
                res.end();
            }
        });
    }
    _validateRequest(req) {
        if (this._methodMetadata.request &&
            this._methodMetadata.request.required &&
            req.body == null) {
            throw http_errors_1.default(http_status_codes_1.default.BAD_REQUEST, "A body is required.");
        }
        if (!this._requestValidator(req.body)) {
            const errorMessage = ajv_utils_1.getValidatorError(this._requestValidator, "does not match required schema", "body");
            throw http_errors_1.default(http_status_codes_1.default.BAD_REQUEST, errorMessage);
        }
    }
    _validateResponse(statusCode, result) {
        const responseValidator = this._responseValidators[statusCode];
        if (!responseValidator) {
            return;
        }
        if (!responseValidator(result)) {
            const errorMessage = ajv_utils_1.getValidatorError(responseValidator, "Response did not match responseSchema.", "response");
            // Throw a real error so it can be logged.
            //  Express will return an Internal Server Error in response.
            throw new Error(errorMessage);
        }
    }
    _collectMethodArgs(req, res) {
        const { handlerArgs } = this._methodMetadata;
        if (!handlerArgs) {
            return [];
        }
        return handlerArgs.map((argMetadata) => this._collectArg(req, res, argMetadata));
    }
    _collectArg(req, res, argMetadata) {
        switch (argMetadata.type) {
            case "body":
                return req.body;
            case "pathParam": {
                return this._collectPathParam(req, argMetadata);
            }
            case "queryParam": {
                return this._collectQueryParam(req, argMetadata);
            }
            case "request": {
                return req;
            }
            case "response": {
                return res;
            }
            default:
                return undefined;
        }
    }
    _collectPathParam(req, argMetadata) {
        const { paramName } = argMetadata;
        const validator = this._pathValidators[paramName];
        let value = req.params[paramName];
        if (validator) {
            value = validator(value);
        }
        return value;
    }
    _collectQueryParam(req, argMetadata) {
        const { paramName } = argMetadata;
        const validator = this._queryValidators[paramName];
        let value = req.query[paramName];
        if (validator) {
            value = validator(value);
        }
        return value;
    }
}
exports.MethodHandler = MethodHandler;
function createQueryParamValidator(metadata, key) {
    // We need to pass the data by object, for type coersion and default values, and requiredness.
    const validate = ajv.compile({
        type: "object",
        properties: {
            value: Object.assign(Object.assign({}, metadata.schema), { 
                // Remove our required: true/false value, as it is not standard json-schema.
                required: undefined }),
        },
        required: metadata.required ? ["value"] : [],
    });
    return (value) => {
        const data = { value };
        if (metadata.required && value === undefined) {
            throw http_errors_1.default(http_status_codes_1.default.BAD_REQUEST, `Query parameter ${key} is required.`);
        }
        if (!validate(data)) {
            throw http_errors_1.default(http_status_codes_1.default.UNPROCESSABLE_ENTITY, ajv_utils_1.getValidatorError(validate, `Query parameter ${key} is invalid.`, key));
        }
        // Data may have been coerced by ajv.
        return data.value;
    };
}
function createPathParamValidator(metadata) {
    // We need to pass the data by object, for type coersion and default values, and requiredness.
    const validate = ajv.compile({
        type: "object",
        properties: {
            value: metadata.schema || {},
        },
        required: metadata.required ? ["value"] : [],
    });
    return (value) => {
        const data = { value };
        if (!validate(data)) {
            // We could pass the validation error here, but a malformed path is typically 404d.
            throw http_errors_1.default(http_status_codes_1.default.NOT_FOUND);
        }
        // Data may have been coerced by ajv.
        return data.value;
    };
}
//# sourceMappingURL=method-handler.js.map