"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.createSwaggerPaths = void 0;
const metadata_1 = require("../metadata");
const controller_utils_1 = require("../controller-utils");
function createSwaggerPaths(...controllers) {
    const result = {};
    for (const controller of controllers) {
        const controllerPaths = createSwaggerControllerPaths(controller);
        Object.assign(result, controllerPaths);
    }
    return result;
}
exports.createSwaggerPaths = createSwaggerPaths;
function createSwaggerControllerPaths(controller) {
    const controllerMetadata = metadata_1.getControllerMetadata(controller.constructor);
    if (!controllerMetadata) {
        throw new Error(`Controller "${controller.constructor.name}" does not have a @controller annotation.`);
    }
    const methods = controller_utils_1.getControllerMethods(controller);
    const result = {};
    for (const { path, metadata } of methods) {
        const key = expressPathToSwaggerPath(path);
        if (!result[key]) {
            result[key] = {};
        }
        result[key][metadata.method.toLowerCase()] = createSwaggerMethodConfig(metadata);
    }
    return result;
}
function createSwaggerMethodConfig(methodMetadata) {
    if (methodMetadata.swaggerOverride) {
        return methodMetadata.swaggerOverride;
    }
    return {
        summary: methodMetadata.summary,
        description: methodMetadata.description,
        tags: methodMetadata.tags,
        parameters: [
            ...createSwaggerMethodPathParameters(methodMetadata),
            ...createSwaggerMethodQueryParameters(methodMetadata)
        ],
        requestBody: createSwaggerRequestBody(methodMetadata),
        responses: createSwaggerResponses(methodMetadata)
    };
}
function createSwaggerRequestBody(methodMetadata) {
    if (!methodMetadata.request) {
        return undefined;
    }
    return {
        required: true,
        content: {
            "application/json": {
                required: methodMetadata.request.required,
                schema: methodMetadata.request.schema
            }
        }
    };
}
function createSwaggerResponses(methodMetadata) {
    const responses = {};
    const keys = Object.keys(methodMetadata.responses || {})
        .map(Number)
        .filter(x => !isNaN(x));
    if (keys.length === 0) {
        return undefined;
    }
    for (const statusCode of keys) {
        const data = methodMetadata.responses[statusCode];
        responses[statusCode] = {
            description: data.description,
            content: {
                "application/json": {
                    schema: data.schema
                }
            }
        };
    }
    return responses;
}
function createSwaggerMethodPathParameters(methodMetadata) {
    const params = [];
    for (const key of Object.keys(methodMetadata.pathParams || {})) {
        const param = methodMetadata.pathParams[key];
        params.push({
            name: key,
            in: "path",
            required: true,
            description: param.description,
            schema: param
        });
    }
    return params;
}
function createSwaggerMethodQueryParameters(methodMetadata) {
    const params = [];
    for (const key of Object.keys(methodMetadata.queryParams || {})) {
        const param = methodMetadata.queryParams[key];
        params.push({
            name: key,
            in: "query",
            required: param.required,
            description: param.description,
            schema: omit(param, "required")
        });
    }
    return params;
}
function expressPathToSwaggerPath(path) {
    function expressToSwagger(fragment) {
        if (fragment.startsWith(":")) {
            return "{" + fragment.substr(1) + "}";
        }
        return fragment;
    }
    return path
        .split("/")
        .map(expressToSwagger)
        .join("/");
}
function omit(obj, ...props) {
    const result = {};
    for (const key of Object.keys(obj)) {
        if (props.includes(key)) {
            continue;
        }
        result[key] = obj[key];
    }
    return result;
}
//# sourceMappingURL=swagger-path-factory.js.map